/*
 * Copyright (c) 1994
 *
 * Bettina Schnor,  Stefan Stille,  Henrik Carlsson
 *
 * TU Braunschweig, Germany
 * Institute for Operating Systems and Computer Networks
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that this copyright
 * notice appears in all copies.  The University of Braunschweig
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/*
 * $Id: system.h,v 1.5 1994/03/09 14:41:41 carlsson Exp $
 * $Log: system.h,v $
 * Revision 1.5  1994/03/09  14:41:41  carlsson
 * Added the availability "categories" HOST_DOWN, HOST_AVAIL, and
 * HOST_NOT_AVAIL.
 *
 * Revision 1.4  1993/12/20  08:47:02  carlsson
 * Removed all KEY_Specs... definitions.
 * Removed the definition of FILE
 * Changed LDAEMON_SERVER to DAEMON_SERVICE
 * Changed LMANAGER_SERVER to LOMA_SERVICE
 * Changed RMANAGER_SERVER to REMA_SERVICE
 * Added definition of NICE
 * Added the message types CONFIG_READ_HOST, CONFIG_READ_SYS, CONFIG_READ_REPLY, KILL_REQUEST and GET_FASTEST_HOST
 *
 * Revision 1.3  1993/12/15  10:04:05  carlsson
 * Moved the definitions of NORMAL_JOB and LONG_JOB to system.h.
 *
 * Revision 1.2  1993/12/06  14:47:18  stille
 * Changed timeout for load-polling second time :-)
 *
 * Revision 1.1  1993/12/06  14:44:08  stille
 * Changed timeout for load-polling
 *
 * Revision 1.0  1993/12/02  15:12:35  stille
 * Initial revision
 *
 * 
 */

#ifndef _SYSDEF_H
#define _SYSDEF_H
/***define your system ****
	WINDOWS NT: WINNT
	WINDOWS 95: WIN95
	LINUX     : LINUX
	SUNOS     : SUNOS
****************************/
#define WIN95
/***************
#define WINNT
#define LINUX
#define SUNOS
***************/ 
#ifdef WIN95
#ifndef WIN32
#define WIN32
#endif
#endif

#ifdef WINNT
#ifndef WIN32
#define WIN32
#endif
#endif

#ifdef LINUX
#ifndef UNIX
#define UNIX
#endif
#endif

#ifdef SUNOS
#ifndef UNIX
#define UNIX
#endif
#endif


#undef  PROTOCOL
#define ROOT_OWNER
#undef  MEASUREMENTS
#undef  USE_AVENRUN
#undef  POLLING 
#define VERBOSE  
#define _LOBA_SYSTEM
#define LOG_DAEMON

#ifdef LOG_DAEMON
#define OPENLOG(prog)   openlog((prog), LOG_PID | LOG_ODELAY, LOG_DAEMON)
#else
#define OPENLOG(prog)   openlog((prog), LOG_PID);
#endif
#define SWAP_AVAIL 20           /* Minimum amount of swap space available at host (in %) */
#define MEM_AVAIL 100           /* Minimum amount of physical memory available (in pages) */
#define LOAD_CHANGE 10.0        /* Percentage of load change for requiring new load broadcast */
#define IDLE_HOST 0.95          /* Limit for considering host idle */
#define LOCAL_REPLY 15          /* Minimum interval in sec between two replies of idle local host */
#define SWAP_VALIDITY 10        /* Load of swap-server is added at this percentage */
#define THRESHOLD 30            /* Percentage by which the load index of selected host must be lower than local index */
#define MINIMUM_TOP 3           /* Minimum number of top-hosts when top-list is to be considered */
#define MAGIC   180568          /* Magic number used for identification of messages */

#define LOAD_POLL 0             /* Message type: poll for load information */
#define LOAD_BROADCAST 1        /* Message type: broadcast of load information */
#define APPLICATION_REQUEST 2   /* Request of application */
#define APPLICATION_REPLY 3     /* Reply to application */
#define APPLICATION_ERROR 4     /* Errormessage to be sent to application */
#define PING 5                  /* Ping load-manager */
#define PING_REPLY 6            /* Reply to ping from load-manager */
#define AUTH_REQUEST 7          /* Request for authentication (from rexec-manager) */
#define AUTH_REPLY 8            /* Reply to authentication request */
#define LOAD_REQUEST 9          /* Request for local load (no broadcast, just send back load index) */
#define GET_ALL_HOSTS 10        /* Get load indexes of all hosts in the system */
#define BROADCAST_FORCE 11      /* Force broadcast of load-index (used when loadmanager comes up) */
#define CONFIG_READ_HOST 12     /* Request for reading host information in configuration file */
#define CONFIG_READ_SYS 13      /* Request for reading system information in configuration file */
#define CONFIG_READ_REPLY 14    /* Reply to read request */
#define KILL_REQUEST 15         /* Request to end YALB system */
#define GET_FASTEST_HOST 16     /* Get fastest host */
#define GET_OPT_HOST    17       /*Get Optimum host*/ 

#define MAXNAMELENGTH 50        /* Maximum name length for machines and keywords in config-file */
#define MAXPARAMS 8             /* Maximum number of parameters per host in config-file */
#define MAXLINELEN 80           /* Maximum length of line in config-file */
#define MAX_DEPENDENCIES 20     /* Maximum number of client- (swap-)server dependencies */

#ifdef UNIX
#ifdef ROOT_OWNER
#define YALB_PATH "/tmp/YALB"           /* Path for YALB-Data */ 
#else
#define YALB_PATH "/tmp"           /* Path for YALB-Data */ 
#endif
#endif

#ifdef WIN32
#ifdef ROOT_OWNER
#define YALB_PATH "\\tmp\\YALB"           /* Path for YALB-Data */ 
#else
#define YALB_PATH "\\tmp"           /* Path for YALB-Data */ 
#endif
#define LAMANAGER "lmanager"
#define RMANAGER "remote"
#define YALB_DAEMON "\\usr\\local\\etc"
#define LDAEMON_BASE "ldaemon"
#endif

#define SHARED "yalb.shared"       /* File name for shared memory segment */
#define UNIX_SOCKET "yalb.socket"  /* Name of Unix domain socket */
#define UPDATE 1                   /* load update interval in secs. */
/*changed on 3.9.99 #define UPDATE 5               /* load update interval in secs. */
#define MAX_MISSING_UPDATES 10  /* Maximum number of missing load updates 
				   for still beeing considered as up. */
#define SMOOTHING 0.84648172    /* Value for exponential smoothing of 
				 * runqueue etc. calculated as:
				 * exp( -UPDATE/SEC) where SEC is
				 * the smoothing interval in sec */
#ifndef BROADCAST_MASK
#define BROADCAST_MASK 0xffffff00    /* Broadcast mask */  
#endif /* ! BROADCAST_MASK */
#define ENDOFHOSTLIST 0xffffffff /* Terminating entry of hostlist */
#ifdef POLLING
#define TIMEOUT 80000           /* Timeout-interval for polling in *microseconds* */
#else
#define TIMEOUT 200000          /* Timeout-interval for polling in *microseconds* */
#endif /* POLLING */
#define SYSTEM 2                /* Index for system-time */
#define MAXLENGTH 4096          /* Maximum length of data per message */
/** XXX (schoenfr): fix me: a simple 64 for MAXHOSTS gives 
 ** message too long :-( **/
#define MAXHOSTS 48           /* Maximum number of hosts supported by system */
#define ATTEMPTS 10            /* Number of tries to get a reply from load-manager */
#define OFFSET 100 /*1               /* Number of seconds a call to load-manager should last */

#define MAX_PATH_LEN 1024       /* Maximum path length */

#define DAEMON_SERVICE "ldaemon" /* Name of loaddaemon server in /etc/services */
#define LOMA_SERVICE "lmanager" /* Name of loadmanager server in /etc/services */
#define REMA_SERVICE "remote"  /* Name of rexecmanager server in /etc/services */ 
#ifdef WIN32
#define SIG_CONFIG_READ SIGTERM
#else
#define SIG_CONFIG_READ SIGUSR1 /* Signal for forcing loadmanager to reread configuration file */
#endif
#define TRUE 1
#define FALSE 0

/*
 * Nice value for long jobs
 */

#define NICE 10


/*
 * Kind of job to be performed
 */

#define NORMAL_JOB 0
#define LONG_JOB 1

/*
 * Error-codes
 */

#define INVALID_NUMBER 0        /* Invalid number of hosts in request from application */

/*
 * Message types
 */

#define LOAD_POLL 0             /* Poll for load information */
#define LOAD_BROADCAST 1        /* Load broadcast */

#define HOST_DOWN -2           /* The host is down */
#define HOST_NOT_AVAIL -1      /* The host is not available */
#define HOST_AVAIL 0           /* The host is available */

#endif /* _SYSDEF_H*/
